/*
 * description : DTPML-485 example code. 
 * version : 1.0
 * Board : Raspberry 2 Model B V 1.1
 * Support Module : DTPML-485-81, DTPML-485-151
 *
 * Further information:
 *   https://projects.drogon.net/raspberry-pi/wiringpi/
 *   https://www.diwellshop.com/
 *
 * license : free.
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY.
 * Revision history.
 * 1. 2020.01.02 : First version is released.
 ***********************************************************************
 */

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <wiringPi.h>
#include <wiringSerial.h>
#include "desc.h"

#define DE_RE	1  			// 485 IC DE/RE GPIO ( wPi:1 = BCM:18 = physical:12 )
#define BUFFER_SIZE 8			// RX max buffer
#define REQUEST_CYCLE 	50		// 50ms : Request cycle. Should be at least 50ms
#define TIMEOUT	20			// 20ms : Note. Do not modify this value!!!
#define TRANSMISSION_TIME	5	// 5ms : Note. Do not modify this value!!!
#define DE_RE_CONTROL_TIME	1	// 1ms : Note. Do not modify this value!!!

modbus_t telegram;			// Master query structure

uint8_t *ptr, com_result;
uint8_t RX_BUF[BUFFER_SIZE], rx_count;
uint16_t au16data[2];
uint16_t Request_interval = REQUEST_CYCLE-TIMEOUT-TRANSMISSION_TIME-DE_RE_CONTROL_TIME; // 50ms-20ms-5ms-1ms = 24ms
int16_t Object, Sensor;
int32_t fd;
uint32_t Start_ms, request_count=0;

int main ()
{	
   	telegram.u8id = 1;                               // slave ID 
   	telegram.u8fct = 3;                              // function code = 03 (Read Holding Registers) 
   	telegram.u16RegAdd = 1200;                       // start address in slave : Fixed value(1200)
	telegram.u16RegNo = 2;                           // number of elements to read : Fixed value(2)
	telegram.au16reg = au16data;                     // pointer to a memory array in the raspberry 

	if ((fd = serialOpen ("/dev/ttyAMA0", 19200)) < 0)
	{
		fprintf (stderr, "Unable to open serial device: %s\n", strerror (errno)) ;
		return 1 ;
	}
	if (wiringPiSetup () == -1)
	{
		fprintf (stdout, "Unable to start wiringPi: %s\n", strerror (errno)) ;
		return 1 ;
	}
	pinMode(DE_RE, OUTPUT);				// DE/RE PORT Setting : OUTPUT
	INPUT485;					// 485 IC Control : Input Mode.

        delay(200);                   			// waiting for Sensor init.

	while(request_count++ < 100)  			// repeat 100 times.
  	{
		printf ("%d :", request_count) ;
		com_result = query(telegram);		// communication.

    		if(com_result == CRC_OK)
		{
			Object = (int16_t)au16data[0];		// Object Raw Data
			Sensor = (int16_t)au16data[1];		// Sensor Raw Data
			
			printf ("Object: %3.1f  Sensor: %3.1f\n", (float)Object/10, (float)Sensor/10) ;
		}
		else if(com_result == CRC_ERROR)
		{
			printf("CRC Error. Check the line.\n");
		}
		else if(com_result == NO_REPLY)
		{
			//dummy
		}
		
		Start_ms = millis();
		while( millis()-Start_ms < Request_interval); 		// Delay : next request
 	}
 	printf ("================Stopped================\n") ;
  	return 0 ;
}

int8_t query(modbus_t telegram)
{
	uint8_t au8Buffer[8], count, response_status = NO_REPLY;
	uint16_t crc16, *au16regs;

	au16regs = telegram.au16reg;
 	au8Buffer[ ID ]         = telegram.u8id;
    	au8Buffer[ FUNC ]       = telegram.u8fct;
    	au8Buffer[ ADD_HI ]     = highByte(telegram.u16RegAdd );
    	au8Buffer[ ADD_LO ]     = lowByte( telegram.u16RegAdd );
	au8Buffer[ NB_HI ]      = highByte(telegram.u16RegNo );
        au8Buffer[ NB_LO ]      = lowByte( telegram.u16RegNo );
	crc16 = CalcCRC16(au8Buffer, 6);
	au8Buffer[ TX_CRC_HI ]  = lowByte(crc16);   // reverse 
        au8Buffer[ TX_CRC_LO ]  = highByte(crc16);  // reverse

	OUTPUT485;				// 485 IC Output Mode
	delayMicroseconds(1);
	fflush (stdout) ;			// flush buffer
	ptr = au8Buffer;
	rx_count = 0;
	for(count=0; count<8; count++)
	{
		serialPutchar (fd, *ptr++) ; 	// transfer data
	}
	Start_ms = millis();
	while( millis()-Start_ms < TRANSMISSION_TIME); // wait for transfer time.

	INPUT485;				// 485 IC input MODE
	Start_ms = millis();
	while( millis()-Start_ms < TIMEOUT)	
	{
		while (serialDataAvail (fd))
		{ 
			RX_BUF[rx_count] = serialGetchar (fd);
	
			switch(rx_count)
			{
			case 0:
				if(RX_BUF[0] == 0x01){rx_count++;}
				break;
			case 1:
				if(RX_BUF[1] == 0x03){rx_count++;}
				else{rx_count=0;}
				break;
			case 2:
				if(RX_BUF[2] == 0x04){rx_count++;}
				else{rx_count=0;}
				break;
			case 3 ... 7:
				rx_count++;
				break;
			case 8:
				crc16 = CalcCRC16(RX_BUF, rx_count-1);			// CRC calculation.

				if(crc16 == (RX_BUF[rx_count-1]|RX_BUF[rx_count]<<8))	// CRC verification.
				{	
					au16regs[0] = (uint16_t)RX_BUF[3]<<8 |(uint16_t)RX_BUF[4];
					au16regs[1] = (uint16_t)RX_BUF[5]<<8 |(uint16_t)RX_BUF[6];
					response_status = CRC_OK;
				}
				else
				{
					response_status = CRC_ERROR;
				}
				rx_count=0;   
				break;
			default: rx_count=0; break;
	
			} // end switch(rx_count)
		} //end while (serialDataAvail (fd))
	}
	return response_status;
}

uint8_t highByte(uint16_t Data)
{
	return (Data>>8)&0xFF;
}

uint8_t lowByte(uint16_t Data)
{
	return Data&0xFF;
}

uint16_t CalcCRC16(uint8_t *cdata, uint16_t ilen)
{
    uint16_t AccumCRC16 = 0xffff;
    uint8_t i,j;
    for(j=0;j<ilen;j++)
    {
        i= (AccumCRC16 ^ *(cdata++) ) & 255;
        AccumCRC16 = ((AccumCRC16>>8) ^ TableCRC16[i]) & 0xffff;
    }
    return AccumCRC16;
}

